<!DOCTYPE html>
<html>

<head>
    <style>
        table {
            width: 100%;
            border-collapse: collapse;
            margin: 15px 0;
        }

        th,
        td {
            border: 1px solid #999;
            padding: 0.5rem;
            text-align: left;
        }

        th {
            background-color: #f3f3f3;
            color: #333;
        }

        tr:nth-child(even) {
            background-color: #f2f2f2;
        }

        a {
            color: #007BFF;
            text-decoration: none;
        }

        a:hover {
            color: #0056b3;
            text-decoration: none;
        }

        kbd {
            padding: 0.2em 0.4em;
            font-size: 0.87em;
            color: #24292e;
            background-color: #f6f8fa;
            border: 1px solid #d1d5da;
            border-radius: 3px;
            box-shadow: inset 0 -1px 0 #d1d5da;
            font-family: Consolas, "Liberation Mono", Menlo, Courier, monospace;
        }
    </style>
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/default.min.css">
    <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/highlight.min.js"></script>
    <script type="text/x-mathjax-config">
    MathJax.Hub.Config({tex2jax: {inlineMath: [['$','$']]}});
    </script>
    <script id="MathJax-script" async src="https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-mml-chtml.js"></script>
    <title>${title}</title>
</head>

<body>
    ${content}
    <script>hljs.highlightAll();</script>
</body>

</html>
